import angular from 'angular';
import { Injectable } from 'angular-ts-decorators';
import autobind from 'autobind-decorator';
import { ApiServiceType } from '../api/api.enum';
import { ApiService } from '../api/api.interface';
import { Bookmark, BookmarkService } from '../bookmark/bookmark.interface';
import { FailedRestoreDataError } from '../errors/errors';
import { MessageCommand } from '../global-shared.enum';
import { PlatformService } from '../global-shared.interface';
import { LogService } from '../log/log.service';
import { StoreKey } from '../store/store.enum';
import { StoreService } from '../store/store.service';
import { SyncType } from '../sync/sync.enum';
import { UpgradeService } from '../upgrade/upgrade.service';
import { UtilityService } from '../utility/utility.service';
import { AutoBackUpSchedule, Backup, BackupSync } from './backup-restore.interface';

@autobind
@Injectable('BackupRestoreService')
export class BackupRestoreService {
  $q: ng.IQService;
  apiSvc: ApiService;
  bookmarkSvc: BookmarkService;
  logSvc: LogService;
  platformSvc: PlatformService;
  storeSvc: StoreService;
  upgradeSvc: UpgradeService;
  utilitySvc: UtilityService;

  static $inject = [
    '$q',
    'ApiService',
    'BookmarkService',
    'LogService',
    'PlatformService',
    'StoreService',
    'UpgradeService',
    'UtilityService'
  ];
  constructor(
    $q: ng.IQService,
    ApiSvc: ApiService,
    BookmarkSvc: BookmarkService,
    LogSvc: LogService,
    PlatformSvc: PlatformService,
    StoreSvc: StoreService,
    UpgradeSvc: UpgradeService,
    UtilitySvc: UtilityService
  ) {
    this.$q = $q;
    this.apiSvc = ApiSvc;
    this.bookmarkSvc = BookmarkSvc;
    this.logSvc = LogSvc;
    this.platformSvc = PlatformSvc;
    this.storeSvc = StoreSvc;
    this.upgradeSvc = UpgradeSvc;
    this.utilitySvc = UtilitySvc;
  }

  createBackupData(bookmarksData: Bookmark[], syncId: string, serviceUrl: string, syncVersion: string): Backup {
    const backupData: Backup = {
      xbrowsersync: {
        date: this.utilitySvc.getDateTimeString(new Date()),
        sync: {},
        data: {}
      }
    };

    // Add sync info if provided
    if (syncId) {
      backupData.xbrowsersync!.sync = this.createSyncInfoObject(syncId, serviceUrl, syncVersion);
    }

    // Add bookmarks
    backupData.xbrowsersync!.data.bookmarks = bookmarksData;
    return backupData;
  }

  createSyncInfoObject(syncId: string, serviceUrl: string, syncVersion?: string): BackupSync {
    return {
      id: syncId,
      type: ApiServiceType.xBrowserSync,
      url: serviceUrl,
      version: syncVersion
    } as BackupSync;
  }

  getBackupFilename(): string {
    const fileName = `xbs_backup_${this.utilitySvc.getDateTimeString(new Date())}.txt`;
    return fileName;
  }

  getSetAutoBackUpSchedule(newValue?: AutoBackUpSchedule | null): ng.IPromise<AutoBackUpSchedule> {
    if (angular.isUndefined(newValue)) {
      return this.storeSvc.get<AutoBackUpSchedule>(StoreKey.AutoBackUpSchedule);
    }
    return this.storeSvc.set(StoreKey.AutoBackUpSchedule, newValue).then(() => {
      if (newValue === null) {
        this.logSvc.logInfo(`Auto back up schedule cleared`);
      } else {
        this.logSvc.logInfo(
          `Auto back up schedule: ${newValue.autoBackUpHour}:${newValue.autoBackUpMinute} every ${newValue.autoBackUpNumber} ${newValue.autoBackUpUnit}`
        );
      }
      return newValue;
    });
  }

  restoreBackupData(backupData: Backup): ng.IPromise<void> {
    let bookmarksToRestore: Bookmark[];
    let serviceUrl: string;
    let syncEnabled: boolean;
    let syncId: string;
    let syncVersion: string;

    switch (true) {
      case !angular.isUndefined(backupData.xbrowsersync): // v1.5.0+
        bookmarksToRestore = backupData.xbrowsersync.data?.bookmarks;
        serviceUrl = backupData.xbrowsersync.sync?.url;
        syncId = backupData.xbrowsersync.sync?.id;
        syncVersion = backupData.xbrowsersync.sync?.version;
        break;
      case !angular.isUndefined(backupData.xBrowserSync): // Prior to v1.5.0
        bookmarksToRestore = backupData.xBrowserSync.bookmarks;
        syncId = backupData.xBrowserSync.id;
        break;
      default:
        // Invalid restore data
        throw new FailedRestoreDataError();
    }

    this.logSvc.logInfo('Restoring data');

    // Upgrade the bookmarks if required
    return this.platformSvc
      .getAppVersion()
      .then((currentVersion) => {
        return syncVersion === currentVersion
          ? bookmarksToRestore
          : this.upgradeSvc.upgradeBookmarks(currentVersion, syncVersion, bookmarksToRestore);
      })
      .then((upgradedBookmarks) =>
        this.utilitySvc
          .isSyncEnabled()
          .then((cachedSyncEnabled) => {
            syncEnabled = cachedSyncEnabled;

            // If synced check service status before starting restore, otherwise restore sync settings
            return syncEnabled
              ? this.apiSvc.checkServiceStatus()
              : this.$q((resolve, reject) => {
                  // Clear current password and set sync ID if supplied
                  this.$q
                    .all([
                      this.storeSvc.remove(StoreKey.Password),
                      syncId ? this.storeSvc.set(StoreKey.SyncId, syncId) : this.$q.resolve(),
                      serviceUrl ? this.utilitySvc.updateServiceUrl(serviceUrl) : this.$q.resolve()
                    ])
                    .then(resolve)
                    .catch(reject);
                });
          })
          // Start restore
          .then(() =>
            this.platformSvc.queueSync(
              {
                bookmarks: upgradedBookmarks,
                type: !syncEnabled ? SyncType.Local : SyncType.LocalAndRemote
              },
              MessageCommand.RestoreBookmarks
            )
          )
      )
      .then(() => {});
  }

  runAutoBackUp(): void {
    this.logSvc.logInfo('Running auto back up');
    this.saveBackupFile(false);
  }

  saveBackupFile(displaySaveDialog?: boolean): ng.IPromise<string | void> {
    let filename: string;
    return this.$q
      .all([
        this.bookmarkSvc.getBookmarksForExport(),
        this.storeSvc.get<string>(StoreKey.SyncId),
        this.utilitySvc.getServiceUrl(),
        this.utilitySvc.getSyncVersion(),
        this.utilitySvc.isSyncEnabled()
      ])
      .then((data) => {
        const [bookmarksData, syncId, serviceUrl, syncVersion, syncEnabled] = data;
        const backupData = this.createBackupData(
          bookmarksData,
          syncEnabled ? syncId : undefined,
          syncEnabled ? serviceUrl : undefined,
          syncEnabled ? syncVersion : undefined
        );

        // Beautify json and download data
        const beautifiedJson = JSON.stringify(backupData, null, 2);
        filename = this.getBackupFilename();
        return this.platformSvc.downloadFile(filename, beautifiedJson, displaySaveDialog);
      });
  }
}
